/********************************************************************************
    Copyright (C) 1999  Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include <iostream.h>
#include <fstream.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "input/streamsrc_istr.hh"
#include "system/sysdemux.hh"
#include "system/videoauconv.hh"
#include "system/ptsvideo.hh"
#include "system/sysdec1.hh"
#include "system/resync.hh"
#include "system/syncer.hh"
#include "vdecoder.hh"
#include "output/accumulator.hh"
#include "output/out_x11.hh"
#include "ulib/graphics/fileio/write_yuv.hh"
#include "ulib/graphics/color/colorspace.hh"
#include "mpg_dec_imp.h"
 
#include "glb.h"
//#define USE_X11_OUT 1

//static pthread_mutex_t mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t x11_mutex = PTHREAD_MUTEX_INITIALIZER;

//#define TEST2  //Nots: do test in one channel!!!!!
#ifdef TEST2
static int count=0;
static FILE *f;
#endif

static void *Video_Decoding(void *arg)
{
    TMpegDecoderImp *decoder=(TMpegDecoderImp *)arg;
    decoder->Execute();
    return NULL;
}


void TMpegDecoderImp:: Execute( void )
{
fprintf( stderr, "TMpegDecoder:: begin Execute..\n");
    InitStreamSource();
fprintf( stderr, "TMpegDecoder:: end Init Stream Source..\n");
    InitVideoDecoder1();
fprintf( stderr, "TMpegDecoder:: end Init Video 1..\n");
    InitVideoDecoder2();
fprintf( stderr, "[%d] TMpegDecoder:: end Init..\n", getpid());

    while( !end )
    {
        if( output_type==FRAME_OUTPUT_X11 )
        {
#ifdef USE_X11_OUT
            //TiFrameOutput* outputobject=lastoutput;
            VideoOutput* outputobject=(VideoOutput *)lastoutput;
            if (outputobject->PictureAvailable())
            {
                if (outputobject->AskPTSOfNextToBeDisplayed())
                   syncer->WaitUntil(outputobject->AskPTSOfNextToBeDisplayed());
pthread_mutex_lock( &x11_mutex );
                outputobject->ShowPicture();
pthread_mutex_unlock( &x11_mutex );
            }
            else
            {
                get_next_frame();
            }
#endif
        }
        else if( output_type==FRAME_OUTPUT_PIPE )
            get_next_frame();
    }
fprintf( stderr, "MpegDecoder exit!!!\n");
}

void TMpegDecoderImp:: Terminate( void )
{
    end=true;
    pthread_join(vd_thread, NULL);
fprintf( stderr, "end Terminate vd_thread\n");
}

TMpegDecoder:: TMpegDecoder( int i_fd, int o_fd, bool i_frame_only, TOutputType type )
{
    impl= (void *)(new TMpegDecoderImp( i_fd, o_fd, i_frame_only, type ));
}

void TMpegDecoder:: run (void )
{
    ((TMpegDecoderImp *)impl)->run();
}

void TMpegDecoder:: Terminate (void )
{
    ((TMpegDecoderImp *)impl)->Terminate();
}

TMpegDecoderImp:: TMpegDecoderImp( int i_fd, int o_fd, bool i_frame_only, TOutputType type )
{
    end=false;
    mpeg_fd=i_fd;
    iframe_fd=o_fd;
    system_dec= new SystemDecoder_MPEG1;
    system_resync= new SystemResyncer;
    system_demux= new SystemDemux(*system_resync);
    video_extractor= new SystemDemuxExtractor;
    vidcon = new VideoAUConverter(*video_extractor);
    video_ptsvideo= new PTSVideoCompleter(*vidcon);
    video_dec= new VideoDecoder(*video_ptsvideo);
    syncer = new Syncer_Realtime;
    static VideoDecoder_Options vdec_options;
    video_dec->SetOptions(vdec_options);
    if( !i_frame_only )
        video_dec->enable_b_p_frame();
    output_type=type;
}

void TMpegDecoderImp:: run( void )
{
    pthread_create(&vd_thread,NULL, Video_Decoding, (void *)this);
}

void TMpegDecoderImp:: get_next_frame( void )
{
//pthread_mutex_lock( &mutex );
    video_dec->DecodeAFrame();
//pthread_mutex_unlock( &mutex );
}
 
void TMpegDecoderImp:: InitStreamSource( void )
{
  input_stream= new ifstream(mpeg_fd);  // standard in
  strsrc_istr= new StreamSource_IStream;
  strsrc_istr->SetIStream(*input_stream);
  streamsrc=strsrc_istr;
}

void TMpegDecoderImp:: InitVideoDecoder1()
{
  system_dec->SetSource(*streamsrc);
fprintf( stderr, "1\n");
  system_resync->SetSource(system_dec);
fprintf( stderr, "2\n");

  system_demux->EnableChannel(StreamIDBase_Video|0);
fprintf( stderr, "3\n");

  video_extractor->SetSource(system_demux);
fprintf( stderr, "4\n");
  video_extractor->SetChannel(StreamIDBase_Video|0);
}

void TMpegDecoderImp:: InitVideoDecoder2( void )
{
    if( output_type==FRAME_OUTPUT_PIPE )
    {
        lastoutput= new TiFrameOutput(iframe_fd);
    }
    else
    {
#ifdef USE_X11_OUT
pthread_mutex_lock( &x11_mutex );
fprintf( stderr, "create VideoSink_X11\n");
        lastoutput= new VideoSink_X11;
fprintf( stderr, "end VideoSink_X11\n");
pthread_mutex_unlock( &x11_mutex );
#endif
    }
    video_dec->SetSink(*lastoutput);
}

// : first(true), pic_available(false)
TiFrameOutput:: TiFrameOutput(int fd)
{
    iframe_fd=fd;
    first=true;
    pic_available=false;
}

void TiFrameOutput:: BeginPicture(const DecodedImageData *dimg )
{
    StartAccumulation(0,dimg->m_height-1,false);
#ifdef TEST2
    if( count>20 )return;
    char s[128];
    sprintf(s,"/tmp/b%d.raw", count );
    f=fopen(s, "wb");
    count++;
#endif
}

void TiFrameOutput:: ShowMBRows(DecodedImageData *dimg )
{
  nextpts = dimg->m_timing.pts;

  if (first)
    {
#ifdef IMG_OUT_YUV
      ostr.attach(iframe_fd);
      writer.SetYUVStream(ostr);
#else
      writer.SetRGBStream(iframe_fd);
#endif
      first=false;
    }

  DecodedImageData* decimg = Accumulate(dimg);
  if (!decimg)
    return;

//log_Console( "out iframe to mpeg motion detector");
#ifdef IMG_OUT_YUV
  writer.WriteImage(decimg->m_image);
#else
    Image_RGB<Pixel> rgb_image;
    ImageParam_YUV param;
    decimg->m_image.GetParam(param);
    switch( param.chroma )
    {
        case Chroma444:
log_Console( "YUV4:4:4");
            YUV2RGB_444( decimg->m_image, rgb_image); 
            break;

        case Chroma422:
log_Console( "YUV4:2:2");
            YUV2RGB_422( decimg->m_image, rgb_image); 

        case Chroma420:
//log_Console( "YUV4:2:0");
            YUV2RGB_420( decimg->m_image, rgb_image); 
    }
    writer.WriteImage( rgb_image );
#endif
}

void TiFrameOutput:: FinishedPicture( void )
{
    pic_available=true;
#ifdef TEST2
    fclose(f);
#endif
}

#if 0
main()
{
    TMpegDecoderImp *decoder= new TMpegDecoderImp(0,1);
    unsigned char frame[640*480*3];

    while(1)
    {
        decoder->get_next_frame();
    }

}
#endif


FileWriter_RGB::FileWriter_RGB()
{
}

void FileWriter_RGB::SetRGBStream( int fd )
{
    out_fd=fd;
}

void FileWriter_RGB::WriteImage(const Image_RGB<Pixel>& img )
{
  ImageParam param;
  img.GetParam(param);

  const Pixel*const* R = img.AskFrameR_const();
  const Pixel*const* G = img.AskFrameG_const();
  const Pixel*const* B = img.AskFrameB_const();

  uint8* linebuf = new uint8[param.width*3];

  for (int y=0;y<param.height;y++)
    {
      uint8* p = linebuf;
      for (int x=0;x<param.width;x++)
        {
          // This ugly code allows the compiler to schedule the
          // commands to parallel pipelines.

          uint8 a,b,c;
          a = R[y][x];
          b = G[y][x];
          c = B[y][x];
          *p++ = R[y][x];
          *p++ = G[y][x];
          *p++ = B[y][x];
        }

      write( out_fd, linebuf, param.width*3 );
#ifdef TEST2
    if( count<20  )
    {
        fwrite( linebuf, param.width*3, 1, f );
    }
#endif
    }

  delete[] linebuf;
}
