//*****************************************************************************
//
//	Copyright (c) 2003 ICP Electronics Inc. All Rights Reserved.
//
//	FILE:
//		cs_file_lock.c
//
//	ABSTRACT:
//		File Lock for file critical section
//
//	HISTORY:
//		01/15/03	Ralph Nieh - Add log info.
//		12/23/02	Ralph Nieh - created
//
//*****************************************************************************
#ifndef WIN32

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include "../include/cs_file_lock.h"
//#include "SettingMgrLog.h"

#define CSF_INVALID_ID -1
#define CSF_FAIL -1
#define CSF_PATH "/tmp/"

CSF_LOCK_RET CSF_Lock(char* name, int timeout)//, CSF_LOCK_RET lock)
{
  CSF_LOCK_RET ret;
  int sleep_time = 1;
  int sleep_count = 0;
  int try_to_lock = 1;
  int forever = 0;
  char lock_name[CSF_MAX_NAME_BUFFER] = {0};

  if (!name || strlen(name) == 0)
  {
    ret.status = CSF_LOCK_INVALID_NAME;
    ret.id = CSF_INVALID_ID;
    return ret;
  }
  else
    strcat(strcat(lock_name, CSF_PATH), name);


  if (timeout == 0)
  {
    sleep_time = 1;
    forever = 1;
  }

  while (try_to_lock && (forever || sleep_count < timeout))
  {
    ret.id = open(lock_name, O_RDWR | O_CREAT | O_EXCL, 0444);

    if (ret.id == CSF_INVALID_ID)
    {
		//if (lock.pid == getpid() && lock.id != CSF_INVALID_ID && lock.status == CSF_LOCK_OK)
		//	return lock;
		//SMGR_LOG1("<%d> Cannot Enter File Critical Section\n", getpid());
		//TRACE1("<%d> Cannot Enter File Critical Section\n", getpid());
		sleep(sleep_time);
		sleep_count++;
    }
    else
    {
      try_to_lock = 0;
      //SMGR_LOG1("<%d> File Lock Successfully, Enter File Critical Section\n", getpid());
      //TRACE1("<%d> File Lock Successfully, Enter File Critical Section\n", getpid());
      ret.status = CSF_LOCK_OK;
	  //ret.pid = getpid();
      return ret;
    }
  }

  ret.status = CSF_LOCK_TIMEOUT;
  return ret;
}

CSF_LOCK_RET CSF_Unlock(char* name, int id)
{
	CSF_LOCK_RET ret;

	ret.id = id;

	if (name && CSF_INVALID_ID != id)
	{
		char unlock_name[CSF_MAX_NAME_BUFFER] = {0};

		strcat(strcat(unlock_name, CSF_PATH), name);

		close(id);
		if (0 == unlink(unlock_name))
		{
			//SMGR_LOG1("<%d> File Unlock Successfully, Leave File Critical Section\n", getpid());
			//TRACE1("<%d> File Unlock Successfully, Leave File Critical Section\n", getpid());
			ret.status = CSF_UNLOCK_OK;
			return ret;
		}
	}

	ret.status = CSF_UNLOCK_FAIL;
	return ret;
}
#endif
