/****************************************************************************  
 
  Copyright (c) 1999,2000,2001 WU-FTPD Development Group.  
  All rights reserved.
  
  Portions Copyright (c) 1980, 1985, 1988, 1989, 1990, 1991, 1993, 1994
    The Regents of the University of California.
  Portions Copyright (c) 1993, 1994 Washington University in Saint Louis.
  Portions Copyright (c) 1996, 1998 Berkeley Software Design, Inc.
  Portions Copyright (c) 1989 Massachusetts Institute of Technology.
  Portions Copyright (c) 1998 Sendmail, Inc.
  Portions Copyright (c) 1983, 1995, 1996, 1997 Eric P.  Allman.
  Portions Copyright (c) 1997 by Stan Barber.
  Portions Copyright (c) 1997 by Kent Landfield.
  Portions Copyright (c) 1991, 1992, 1993, 1994, 1995, 1996, 1997
    Free Software Foundation, Inc.  
 
  Use and distribution of this software and its source code are governed 
  by the terms and conditions of the WU-FTPD Software License ("LICENSE").
 
  If you did not receive a copy of the license, it may be obtained online
  at http://www.wu-ftpd.org/license.html.
 
  $Id: domain.c,v 1.17 2001/12/22 09:25:52 wuftpd Exp $
 
****************************************************************************/
/*
 * domain.c  - DNS functions for WU-FTPD using res_*
 *
 * INITIAL AUTHOR - *      Nikos Mouat    <nikm@cyberflunk.com>
 */

#include "config.h"
#include <stdlib.h>
#include "proto.h"

#if defined(HAVE_LIBRESOLV)

#  include <sys/types.h>
#  include <sys/param.h>
#  include <stdio.h>
#  include <string.h>
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  if !defined(NO_DNS)
#    include <arpa/nameser.h>
#    include <resolv.h>

#    if !defined(INT16SZ)
#      define INT16SZ sizeof(u_short)
#    endif /* !defined(INT16SZ) */ 

#    if !defined(INT32SZ)
#      define INT32SZ sizeof(u_long)
#    endif /* !defined(INT32SZ) */ 

#    if !defined(INADDRSZ)
#      define INADDRSZ sizeof(struct in_addr)
#    endif /* !defined(INADDRSZ) */ 

#    if !defined(HFIXEDSZ)
#      define HFIXEDSZ sizeof(HEADER)
#    endif /* !defined(HFIXEDSZ) */ 

#  endif /* !defined(NO_DNS) */ 
#  include "extensions.h"

/* these should go in a new ftpd.h perhaps? config.h doesn't seem appropriate */
/* and there does not appear to be a global include file                      */
#  if !defined(TRUE)
#    define TRUE   1
#  endif /* !defined(TRUE) */ 

#  if !defined(FALSE)
#    define FALSE  !TRUE
#  endif /* !defined(FALSE) */ 

struct t_resolver_options {
    char *token;
    int value;
} resolver_options[] = {

#  if defined(RES_DEBUG)
    {
	"debug", RES_DEBUG
    },
#  endif /* defined(RES_DEBUG) */ 
#  if defined(RES_AAONLY)
    {
	"aaonly", RES_AAONLY
    },
#  endif /* defined(RES_AAONLY) */ 
#  if defined(RES_USEVC)
    {
	"usevc", RES_USEVC
    },
#  endif /* defined(RES_USEVC) */ 
#  if defined(RES_STAYOPEN)
    {
	"stayopen", RES_STAYOPEN
    },
#  endif /* defined(RES_STAYOPEN) */ 
#  if defined(RES_PRIMARY)
    {
	"primary", RES_PRIMARY
    },
#  endif /* defined(RES_PRIMARY) */ 
#  if defined(RES_IGNTC)
    {
	"igntc", RES_IGNTC
    },
#  endif /* defined(RES_IGNTC) */ 
#  if defined(RES_RECURSE)
    {
	"recurse", RES_RECURSE
    },
#  endif /* defined(RES_RECURSE) */ 
#  if defined(RES_DEFNAMES)
    {
	"defnames", RES_DEFNAMES
    },
#  endif /* defined(RES_DEFNAMES) */ 
#  if defined(RES_DNSRCH)
    {
	"dnsrch", RES_DNSRCH
    },
#  endif /* defined(RES_DNSRCH) */ 
#  if defined(RES_INSECURE1)
    {
	"insecure1", RES_INSECURE1
    },
#  endif /* defined(RES_INSECURE1) */ 
#  if defined(RES_INSECURE2)
    {
	"insecure2", RES_INSECURE2
    },
#  endif /* defined(RES_INSECURE2) */ 
#  if defined(RES_NOALIASES)
    {
	"noaliases", RES_NOALIASES
    },
#  endif /* defined(RES_NOALIASES) */ 
#  if defined(RES_USE_INET6)
    {
	"use_inet6", RES_USE_INET6
    },
#  endif /* defined(RES_USE_INET6) */ 
#  if defined(RES_ROTATE)
    {
	"rotate", RES_ROTATE
    },
#  endif /* defined(RES_ROTATE) */ 
#  if defined(RES_NOCHECKNAME)
    {
	"nocheckname", RES_NOCHECKNAME
    },
#  endif /* defined(RES_NOCHECKNAME) */ 
#  if defined(RES_KEEPSIG)
    {
	"keepsig", RES_KEEPSIG
    },
#  endif /* defined(RES_KEEPSIG) */ 
#  if defined(RES_DEFAULT)
    {
	"default", RES_DEFAULT
    },
#  endif /* defined(RES_DEFAULT) */ 
    {
	NULL, 0
    }
};

/* globals */
int resolver_initialized = FALSE;
char *remote_hostname = NULL;
char *remote_address = NULL;
int has_reverse_dns = FALSE;
int has_matching_dns = FALSE;

/* Prototypes */
int check_matching_dns(void);
int lookup_ip(struct SOCKSTORAGE *remote_socket, char **fqdn);
int check_name_for_ip(char *name, struct SOCKSTORAGE *remote_socket);
int initialize_dns(struct SOCKSTORAGE *remote_socket);
int check_reverse_dns(void);

/* types for res_* answers */
#  if !defined(NO_DNS)
typedef union {
    HEADER qb1;
    u_char qb2[PACKETSZ];
} querybuf;

#  endif /* !defined(NO_DNS) */ 

/****************************************************************************
 * lookup_ip()
 *   This routine takes an IP address (in a struct SOCKSTORAGE) and returns
 *   the hostname.
 ***************************************************************************/
int lookup_ip(struct SOCKSTORAGE *remote_socket, char **fqdn)
{
#  if !defined(NO_DNS)
    u_char *msg_end, *msg_ptr;
#    if defined(USE_RES_SEND)
    querybuf question, answer;
#    else /* !(defined(USE_RES_SEND)) */ 
    querybuf answer;
#    endif /* !(defined(USE_RES_SEND)) */ 
    int rc, num_answers, num_query, q_type, q_class, q_ttl, q_len;
    char query_name[MAXDNAME + 1];
    char in_addr[MAXDNAME + 1];
    char *a;
#    if defined(INET6)
    int isv4 = 0;
#    endif /* defined(INET6) */ 

    a = (char *) SOCK_ADDR(*remote_socket);
#    if defined(INET6)
    if (SOCK_FAMILY(*remote_socket) == AF_INET)
	isv4 = 1;
    else if (IN6_IS_ADDR_V4MAPPED(&((struct sockaddr_in6 *)remote_socket)->sin6_addr))
    {
	isv4 = 1;
	a += 12; /* move to the IPv4 part of an IPv4-mapped IPv6 address */
    }
    if (isv4) {
#    endif /* defined(INET6) */ 
#    define UC(b) (((int) b) & 0xff)
	sprintf(in_addr, "%u.%u.%u.%u.in-addr.arpa",
		UC(a[3]), UC(a[2]), UC(a[1]), UC(a[0]));
#    undef UC
#    if defined(INET6)
    }
    else {
	int i;
	char *tmp = in_addr;
	for (i = 15; i >= 0; i--) {
	    sprintf(tmp, "%x.%x.", a[i]&0x0f, (a[i]&0xf0)>>4);
	    tmp += 4;
	}
	sprintf(tmp, "IP6.INT");
    }
#    endif /* defined(INET6) */ 

#    if defined(USE_RES_SEND)
    rc = res_mkquery(QUERY, in_addr, C_IN, T_PTR, NULL, 0, NULL, question.qb2, sizeof(querybuf));

    if (rc < 0) {
	return FALSE;
    }

    rc = res_send(question.qb2, rc, answer.qb2, sizeof(querybuf));
#    else /* !(defined(USE_RES_SEND)) */ 
    rc = res_query(in_addr, C_IN, T_PTR, answer.qb2, sizeof(querybuf));
#    endif /* !(defined(USE_RES_SEND)) */ 

    if (rc < 0) {
	return FALSE;
    }

    msg_end = (u_char *) & answer + rc;

    num_answers = ntohs(answer.qb1.ancount);
    num_query = ntohs(answer.qb1.qdcount);

    if (num_answers < 1) {
	/* No answers mean that this hostname doesn't exist..                 */
	return FALSE;
    }

    msg_ptr = answer.qb2 + HFIXEDSZ;

    /* skip over the query */
    for (; (num_query > 0) && (msg_ptr < msg_end); num_query--) {
	msg_ptr += dn_skipname(msg_ptr, msg_end) + QFIXEDSZ;
    }

    for (; num_answers > 0 && (msg_ptr < msg_end); num_answers--) {

	rc = dn_expand(answer.qb2, msg_end, msg_ptr, (void *) query_name, MAXDNAME);

	/* increment our pointer */
	msg_ptr += rc;

	/* read off the various answer information */
	q_type = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_class = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_ttl = _getlong(msg_ptr);
	msg_ptr += INT32SZ;

	q_len = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	switch (q_type) {
	case T_PTR:
	    /* this is our answer, expand the name, allocate space for it and */
	    /* set the return information                                     */
	    rc = dn_expand(answer.qb2, msg_end, msg_ptr, (void *) query_name, MAXDNAME);

	    if (rc >= 0) {
		*fqdn = (char *) malloc(strlen(query_name) + 1);
		if (*fqdn == NULL)
		    return FALSE;

		strcpy(*fqdn, query_name);
		return TRUE;
	    }
	    break;
	default:
	    /* unknown response type.. keep looking */
	    msg_ptr += q_len;
	}
    }

#  endif /* !defined(NO_DNS) */ 

    return FALSE;
}
/****************************************************************************
 * check_name_for_ip()
 *   This routine checks to see if a given IP address is a valid IP for a
 *   given name.
 ***************************************************************************/
int check_name_for_ip(char *name, struct SOCKSTORAGE *remote_socket)
{
#  if !defined(NO_DNS)
#    if defined(USE_RES_SEND)
    querybuf question, answer;
#    else /* !(defined(USE_RES_SEND)) */ 
    querybuf answer;
#    endif /* !(defined(USE_RES_SEND)) */ 
    u_char *msg_end, *msg_ptr;
    char query_name[MAXDNAME + 1];
    int rc, num_answers, num_query, q_type, q_class, q_ttl, q_len;
    int type = T_A;
    char *qaddr;

    qaddr = (char *) SOCK_ADDR(*remote_socket);

#    if defined(INET6)
    if (SOCK_FAMILY(*remote_socket) == AF_INET6) {
	if (IN6_IS_ADDR_V4MAPPED(&((struct sockaddr_in6 *)remote_socket)->sin6_addr))
	qaddr += 12; /* move to the IPv4 part of an IPv4-mapped IPv6 address */
    else
	type = T_AAAA;
    }
#    endif /* defined(INET6) */ 

#    if defined(USE_RES_SEND)
    rc = res_mkquery(QUERY, name, C_IN, type, NULL, 0, NULL, question.qb2, sizeof(querybuf));

    if (rc < 0) {
	return FALSE;
    }

    rc = res_send(question.qb2, rc, answer.qb2, sizeof(querybuf));
#    else /* !(defined(USE_RES_SEND)) */ 
    rc = res_query(name, C_IN, type, answer.qb2, sizeof(querybuf));
#    endif /* !(defined(USE_RES_SEND)) */ 

    if (rc < 0) {
	return FALSE;
    }

    msg_end = (u_char *) & answer + rc;

    num_answers = ntohs(answer.qb1.ancount);
    num_query = ntohs(answer.qb1.qdcount);

    if (num_answers < 1) {
	/* No answers mean that this hostname doesn't exist..                 */
	return FALSE;
    }

    msg_ptr = answer.qb2 + HFIXEDSZ;

    /* skip over the query */
    for (; (num_query > 0) && (msg_ptr < msg_end); num_query--) {
	msg_ptr += dn_skipname(msg_ptr, msg_end) + QFIXEDSZ;
    }

    for (; (num_answers > 0) && (msg_ptr < msg_end); num_answers--) {
	rc = dn_expand(answer.qb2, msg_end, msg_ptr, (void *) query_name, MAXDNAME);

	msg_ptr += rc;

	q_type = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_class = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_ttl = _getlong(msg_ptr);
	msg_ptr += INT32SZ;

	q_len = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	/* look at the type of response that we recieved. If it's a CNAME then */
	/* we need to find out what the CNAME's IP is.                         */
	switch (q_type) {
	case T_CNAME:
	    /* Got a CNAME - hope that there are other answers with the A */
	    msg_ptr += q_len;
	    break;
	case T_A:
	    if (memcmp(msg_ptr, qaddr, sizeof(struct in_addr)) == 0)
		return TRUE;
	    msg_ptr += q_len;
	    break;
#    if defined(INET6)
	case T_AAAA:
	    if (memcmp(msg_ptr, qaddr, sizeof(struct in6_addr)) == 0)
		return TRUE;
	    msg_ptr += q_len;
	    break;
#    endif /* defined(INET6) */ 
	default:
	    /* what the ?? - we don't expect this response type */
	    return FALSE;
	}
    }

#  endif /* !defined(NO_DNS) */ 

    /* no matching IP's */
    return FALSE;
}

/****************************************************************************
 * initialize_dns()
 *    initialize the DNS subsystem, set resolver options and collect global
 *    variables. The remote socket is passed to this routine so that it
 *    can set DNS variables related to the remote site. 
 ***************************************************************************/
int initialize_dns(struct SOCKSTORAGE *remote_socket)
{
#  if !defined(NO_DNS)

    struct aclmember *entry = NULL;
    char *temp_pointer;

    if (resolver_initialized)
	return TRUE;

    /* check to see if the resolver has been initialized already */
    if ((_res.options & RES_INIT) == 0) {
	if ((res_init()) == -1) {
	    /* failed to initialize the resolver */
	    return FALSE;
	}
    }

    /* load the 'dns resolveroptions X' entries from the config. This will */
    /* change when we have the new config file parsing methods             */
    while (getaclentry("dns", &entry)) {
	if (!ARG0 || !ARG1)
	    continue;
	/* there are other DNS options, we only care about 'resolveroptions' */
	if (!strcasecmp(ARG0, "resolveroptions")) {
	    int arg_count;

	    /* read in the options, and set _res.options appropriately        */
	    for (arg_count = 1; ARG[arg_count] != NULL; arg_count++) {
		int operation = 0;
		char *option;
		int table_index;
		int option_bitvalue;

		if (ARG[arg_count][0] == '-') {
		    operation = -1;	/* want to UNSET option */
		    option = &ARG[arg_count][1];
		}
		else if (ARG[arg_count][0] == '+') {
		    operation = 1;	/* want to SET option */
		    option = &ARG[arg_count][1];
		}

		if (operation == 0) {
		    /* no operation specified, assume they meant + so do a SET */
		    operation = 1;
		    option = &ARG[arg_count][0];
		}

		option_bitvalue = 0;

		/* now lookup option bit value in the resolver_options table */
		for (table_index = 0; resolver_options[table_index].token != NULL;
		     table_index++) {
		    if (!strcasecmp(option, resolver_options[table_index].token)) {
			option_bitvalue = resolver_options[table_index].value;
		    }
		}

		/* make sure that we have a valid operation to perform */
		if (option_bitvalue == 0) {
		    /* nope. keep looking at other args. */
		    continue;
		}


		/* okay, let's do this operation then */
		if (operation < 0) {
		    /* turn *off* option */
		    _res.options &= ~option_bitvalue;
		}
		else {
		    _res.options |= option_bitvalue;
		}
	    }
	}
    }

    /* save the remote address */
    temp_pointer = inet_stop(remote_socket);
    remote_address = (char *) malloc(strlen(temp_pointer) + 1);
    if (remote_address == NULL) {
	/* memory error */
	return FALSE;
    }
    /* size should be identical */
    strcpy(remote_address, temp_pointer);

    /* the old code checks remote_address for 0.0.0.0 and if there's a   */
    /* match, sets the hostname to 'localhost' - so let's do that for    */
    /* compatibility and perhaps unknown resolvers which behave this way */
    if (!strcmp(remote_address, "0.0.0.0")) {
	remote_hostname = "localhost";
    }
    else if (lookup_ip(remote_socket, &temp_pointer)) {
	has_reverse_dns = TRUE;
	remote_hostname = temp_pointer;

	/* ok, we should now have hostname and address based on the real */
	/* IP address. Let's check the forward DNS of remote_hostname    */
	/* and see if we get the same address..                          */
	has_matching_dns = check_name_for_ip(remote_hostname, remote_socket);
    }
    else {
	has_reverse_dns = FALSE;
	has_matching_dns = TRUE;	/* no reverse, nothing to match */
    }

    resolver_initialized = TRUE;

#  endif /* !defined(NO_DNS) */ 

    return TRUE;
}
/****************************************************************************
 * check_reverse_dns()
 ***************************************************************************/
int check_reverse_dns(void)
{
#  if !defined(NO_DNS)

    struct aclmember *entry = NULL;
    int rc = TRUE;

    /* check the config to see if we care */
    while (getaclentry("dns", &entry)) {
	if (!ARG0 || !ARG1)
	    continue;
	if (!strcasecmp(ARG0, "refuse_no_reverse")) {
	    FILE *msg_file;
	    char linebuf[MAXPATHLEN];
	    char outbuf[MAXPATHLEN];
	    int code = 530;
	    char *crptr;

	    /* ok, so configuration is telling us to not allow connections */
	    /* that don't have any reverse DNS                             */
	    if (!has_reverse_dns) {
		/* ok, so we need to kick out this user */

		/* check to see if admin wants to override */
		if (ARG2 && (!strcasecmp(ARG2, "override"))) {
		    /* Administrative override - but display the warning anyway */
		    code = 220;
		}

		msg_file = fopen(ARG1, "r");
		if (msg_file != NULL) {
		    while (fgets(linebuf, sizeof(linebuf), msg_file)) {
			if ((crptr = strchr(linebuf, '\n')) != NULL)
			    *crptr = '\0';
			msg_massage(linebuf, outbuf, sizeof(outbuf));
			lreply(code, "%s", outbuf);
		    }
		    fclose(msg_file);
#    if !defined(NO_SUCKING_NEWLINES)
		    lreply(code, "");
#    endif /* !defined(NO_SUCKING_NEWLINES) */ 
		    if (code == 530) {
			reply(code, "");
			rc = FALSE;
		    }
		    else {
			lreply(code, "Administrative Override. Permission granted.");
			lreply(code, "");
		    }
		}
	    }
	}
    }

    return rc;
#  else /* !(!defined(NO_DNS)) */ 
    return TRUE;
#  endif /* !(!defined(NO_DNS)) */ 
}
/****************************************************************************
 * check_matching_dns()
 ***************************************************************************/
int check_matching_dns(void)
{
#  if !defined(NO_DNS)
    struct aclmember *entry = NULL;
    int rc = TRUE;

    /* check the config to see if we care */
    while (getaclentry("dns", &entry)) {
	if (!ARG0 || !ARG1)
	    continue;
	if (!strcasecmp(ARG0, "refuse_mismatch")) {
	    FILE *msg_file;
	    char linebuf[MAXPATHLEN];
	    char outbuf[MAXPATHLEN];
	    int code = 530;
	    char *crptr;

	    /* ok, so configuration is telling us to not allow connections */
	    /* that don't have any reverse DNS                             */
	    if (!has_matching_dns) {
		/* ok, so we need to kick out this user */

		/* check to see if admin wants to override */
		if (ARG2 && (!strcasecmp(ARG2, "override"))) {
		    /* Administrative override - but display the warning anyway */
		    code = 220;
		}

		msg_file = fopen(ARG1, "r");
		if (msg_file != NULL) {
		    while (fgets(linebuf, sizeof(linebuf), msg_file)) {
			if ((crptr = strchr(linebuf, '\n')) != NULL)
			    *crptr = '\0';
			msg_massage(linebuf, outbuf, sizeof(outbuf));
			lreply(code, "%s", outbuf);
		    }
		    fclose(msg_file);
#    if !defined(NO_SUCKING_NEWLINES)
		    lreply(code, "");
#    endif /* !defined(NO_SUCKING_NEWLINES) */ 
		    if (code == 530) {
			reply(code, "");
			rc = FALSE;
		    }
		    else {
			lreply(code, "Administrative Override. Permission granted.");
			lreply(code, "");
		    }
		}
	    }
	}
    }
    return rc;
#  else /* !(!defined(NO_DNS)) */ 
    return TRUE;
#  endif /* !(!defined(NO_DNS)) */ 
}

#endif /* defined(HAVE_LIBRESOLV) */ 
